import click


@click.command("test")
@click.pass_obj
def cli(environment):
    """Run tests for your application (pytest)"""
    click.echo("test command running!")
    environment.run_in_docker("pytest", {"PALM_TEST": True})



import importlib
from pathlib import Path
from typing import List, Optional, Tuple

import click
from pydantic import BaseModel

from palm.plugin_manager import PluginManager
from palm.utils import run_on_host, run_in_docker

from .code_generator import CodeGenerator
from .palm_config import PalmConfig


class Environment:
    def __init__(self, plugin_manager: PluginManager, palm_config: PalmConfig):
        self.home = Path.cwd()
        self.palm = palm_config
        self.plugin_manager = plugin_manager

    def run_in_docker(
        self,
        cmd: str,
        env_vars: Optional[dict] = {},
        no_bin_bash: Optional[bool] = False,
        silent: Optional[bool] = False,
    ) -> Tuple[bool, str]:
        env_vars_list = self._build_env_vars(env_vars)
        return run_in_docker(
            cmd, self.palm.image_name, env_vars_list, no_bin_bash, silent
        )

    def run_on_host(
        self,
        cmd: str,
        check: Optional[bool] = False,
        capture_output: Optional[bool] = False,
    ) -> Tuple[int, str, str]:
        """context wrapper for :obj:`palm.utils.run_on_host`"""
        return run_on_host(cmd, check, capture_output)

    def import_module(self, module_name: str, module_path: Path):
        """Imports a module from a path

        This is useful if you need to import a python module to a command
        defined in your local project, unfortunately normal relative `import`
        statements don't work in that context because palm is running in your
        python path, so you need to use this instead.

        Args:
            module_name (str): Name of the module
            module_path (Path): Absolute Path to the module file

        Returns:
            ModuleSpec: The module spec of the imported module,
            or None if the module was not found
        """
        try:
            spec = importlib.util.spec_from_file_location(module_name, module_path)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
        except ImportError as error:
            click.secho(f"Import error: {error}", fg="red")
            return None
        return mod

    def generate(
        self, template_path: Path, target_path: Path, replacements: dict
    ) -> str:
        """Runs a template through the code generator

        Args:
            template_path (Path): Path to the directory containing the template
            target_path (Path): Path to the directory where the generated code will be written
            replacements (dict): Dict of replacements to make in the template

        Returns:
            str: The path to the generated code
        """
        return CodeGenerator(template_path, target_path, replacements).run()

    def _build_env_vars(self, env_vars: dict) -> List[str]:
        env_vars_list = []
        for key in env_vars.keys():
            env_vars_list.append(f"-e {key.upper()}={env_vars[key]}")
        return env_vars_list

    def plugin_config(self, plugin_name: str) -> Optional[BaseModel]:
        """Returns the config for a plugin

        Args:
            plugin_name (str): The name of the plugin

        Returns:
            Optional[BaseModel]: The config for the plugin, or None if the plugin
            is not found or does not have a config
        """
        plugin = self.plugin_manager.plugins.get(plugin_name)
        if plugin and plugin.config:
            return plugin.config.get()
        return None



from distutils.core import setup
from pathlib import Path
import sys

# require version of setuptools that supports find_namespace_packages
from setuptools import setup

try:
    from setuptools import find_namespace_packages
except ImportError:
    # the user has a downlevel version of setuptools.
    print('Error: palm requires setuptools v40.1.0 or higher.')
    print(
        'Please upgrade setuptools with "pip install --upgrade setuptools" and try again'
    )
    sys.exit(1)

# User-friendly description from README.md
this_directory = Path(__file__).parent
long_description = Path(this_directory, 'README.md').read_text()

setup(
    name='palm-{{plugin_name}}',
    version='0.0.1',
    description='{{plugin_name}} extension for Palm CLI',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='{{author}}',
    author_email='{{author_email}}',
    url='',
    packages=find_namespace_packages(include=['palm', 'palm.*']),
    package_data={'': ['*.yaml', '*.yml']},
    install_requires=['palm>=2.0.0'],
    license='',
    classifiers=[],
)



import re
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Optional

import click

import palm.project_setup_utils as psu
from palm.palm_exceptions import AbortPalm


class Containerizer(ABC):
    """Abstract base class for containerizer classes

    Containerizer.run() is the main entry point for a containerizer.
    Implementations of run() should call check_setup() and generate()
    """

    def __init__(self, ctx, template_dir: Path) -> None:
        """Containerizer constructor

        Args:
            ctx (click.context): The click context object from the calling command
            template_dir (Path): Path to the templates directory
        """
        self.ctx = ctx
        # By default, image name is read from .palm/config.yaml since this is used
        # to execute commands in the container. This may be overridden in a containerizer
        # but you would need to ensure the image_name matches in config.yaml
        self.project_name = ctx.obj.palm.image_name
        self.template_dir = template_dir

    @abstractmethod
    def run(self) -> None:
        """Entrypoint for containerization

        Must call check_setup() and generate()
        """
        pass

    def generate(self, target_dir: Path, replacements: Dict) -> None:
        """Generate files for containerization

        Args:
            target_dir (Path): Path to directory where generated files should be placed
            replacements (dict): Dict of replacements to be made in template files
        """
        self.ctx.obj.generate(self.template_dir, target_dir, replacements)
        psu.make_executable(Path(target_dir, "scripts", "entrypoint.sh"))

    def check_setup(self) -> None:
        """Check if containerization is possible

        Raises:
            AbortPalm: If the project is already containerized, or has no .env file
        """
        if self.has_containerization():
            raise AbortPalm("Containerization already exists")

        if not self.has_env():
            try:
                psu.optionally_create_env()
            except AbortPalm:
                raise AbortPalm("Aborting Containerization")

    def has_containerization(self) -> bool:
        """Check whether the project has containerization

        Returns:
            bool: true if Dockerfile or docker-compose.yaml exists
        """
        containerization_files = ["docker-compose.yaml", "Dockerfile"]

        for file in containerization_files:
            if Path(file).exists():
                return True

        return False

    def has_env(self) -> bool:
        """Check whether the project has an .env file

        Returns:
            bool: true if .env file exists
        """
        return psu.has_env()


class PythonContainerizer(Containerizer):
    """Containerizer for Python projects"""

    def __init__(
        self, ctx, template_dir: Path, python_version: Optional[str] = "3.8"
    ) -> None:
        """PythonContainerizer constructor

        Args:
            ctx (click.context): The click context object from the calling command
            template_dir (Path): Path to the templates directory
        """
        self.ctx = ctx
        self.project_name = ctx.obj.palm.image_name
        self.template_dir = template_dir
        self.python_version = python_version
        self.package_manager = ""

    def run(self) -> None:
        """Run the containerizer"""
        self.check_setup()
        self.package_manager = self.detect_package_manager()

        super().generate(self.target_dir, self.replacements)

    @property
    def replacements(self) -> Dict:
        """Dict of replacements to be made in template files"""
        return {
            "project_name": self.project_name,
            "package_manager": self.package_manager,
            "python_version": self.python_version,
        }

    @property
    def target_dir(self) -> Path:
        """Get the target directory for containerization

        Returns:
            Path: Path to the target directory
        """
        return Path.cwd()

    def check_setup(self) -> None:
        if not self.validate_python_version():
            click.secho(f"Invalid python version: {self.python_version}", fg="red")
            return

        try:
            super().check_setup()
        except AbortPalm as e:
            click.secho(str(e), fg="red")
            return

    def detect_package_manager(self) -> str:
        """Determine which package manager is in use,
        supports pip3 and poetry, returns "unknown" if neither are found

        Returns:
            str: pip3 | poetry | unknown
        """
        if self.has_requirements_txt():
            return "pip3"
        if self.has_poetry():
            return "poetry"
        # Unknown package manager, prompt to setup requirements.txt
        try:
            self.optionally_add_requirements_txt()
        except AbortPalm:
            click.secho("Aborting containerization", fg="red")
            sys.exit(1)
        return "pip3"

    def has_requirements_txt(self) -> bool:
        """Check for a requirements.txt file in the project root"""
        return Path("requirements.txt").exists()

    def has_poetry(self) -> bool:
        """Check for a poetry.lock file in the project root"""
        return Path("poetry.lock").exists()

    def optionally_add_requirements_txt(self):
        """Optionally, add a requirements.txt file to the project root if it doesn't exist

        Raises:
            AbortPalm: Abort if user does not want to add requirements.txt
        """
        use_pip = click.confirm(
            "Unable to detect python package manger, requirements.txt will be used by default. Continue?"
        )
        if use_pip:
            Path("requirements.txt").touch()
        else:
            raise AbortPalm("Aborting")

    def validate_python_version(self) -> bool:
        """Validate the python version provided by the user

        Args:
            python_version (str): The python version to validate

        Returns:
            bool: True if the python version is valid
        """
        match = re.match(r"^3\.[0-9]{1,2}$", self.python_version)
        return bool(match)



import importlib
from typing import List

from click import secho

from .plugins.base import BasePlugin as Plugin


class PluginManager:
    def __init__(self) -> None:
        self.plugins = {}
        self.plugin_command_dict = {}

    def load_plugins(self, plugins: List) -> None:
        """Loads a list of plugins, typically from palm config

        Args:
            plugins (list): list of plugin names to load
        """
        for plugin in plugins:
            self.load_plugin(plugin)

    def load_plugin(self, plugin_name: str) -> Plugin:
        """Load a single plugin by name

        Args:
            plugin_name (str): name of the plugin to be loaded

        Raises:
            KeyError: Raised if the requested plugin is not available

        Returns:
            Plugin: Plugin instance
        """
        try:
            module: Plugin = importlib.import_module("." + plugin_name, "palm.plugins")
        except ModuleNotFoundError as e:
            if e.name == "palm.plugins." + plugin_name:
                secho(f"Could not find plugin: {plugin_name}!", fg="red")
            secho(f"Error importing plugin: {e}", fg="red")
            raise

        plugin = module.Plugin
        self.plugins[plugin_name] = plugin
        self.extend_plugin_command_mapping(plugin_name)

        return plugin

    def extend_plugin_command_mapping(self, plugin_name: str) -> None:
        """Merges the plugin commands to the PluginManager plugin_command_dict
        Note that this is using dict merging. In the event of 2 plugins having
        the same command name, the second command will supercede the first,
        replacing it.

        Args:
            plugin_name (str): name of the plugin
        """
        self.plugin_command_dict = {
            **self.plugin_command_dict,
            **self.plugins[plugin_name].command_map(),
        }

    def is_plugin_command(self, command_name: str) -> bool:
        """Check whether a given command name comes from a plugin

        Args:
            command_name (str): Name of the palm command

        Returns:
            bool: True if the command name is registered by a plugin
        """
        return command_name in self.plugin_command_list

    def command_spec(self, command_name) -> importlib.machinery.ModuleSpec:
        """Get the executable module spec for the plugin command

        Args:
            command_name (str): Name of the palm command

        Returns:
            ModuleSpec: ModuleSpec of the command being executed
        """
        plugin_name = self.plugin_command_dict[command_name]
        plugin = self.plugins[plugin_name]
        return plugin.get_command(command_name)

    @property
    def plugin_command_list(self) -> List:
        """Get all commands for installed plugins

        Returns:
            list: List of all commands for installed plugins
        """
        return list(self.plugin_command_dict.keys())



from pathlib import Path

import click

from palm.containerizer import PythonContainerizer


@click.command("containerize")
@click.option(
    "--version",
    multiple=False,
    default="3.8",
    help="Python version to use (default 3.8)",
)
@click.pass_context
def cli(ctx, version: str):
    all_templates_dir = Path(Path(__file__).parents[1], "templates")
    template_dir = all_templates_dir / "containerize"
    PythonContainerizer(ctx, template_dir, version).run()
    click.secho(f"Containerized {ctx.obj.palm.image_name}", fg="green")



from pathlib import Path
from palm.plugins.base import BasePlugin
import pkg_resources

def get_version():
    try:
        version = pkg_resources.require("palm-{{plugin_name}}")[0].version
    except pkg_resources.DistributionNotFound:
        version = 'unknown'
    return version


{{plugin_class_name}} = BasePlugin(
    name = '{{plugin_name}}', 
    command_dir = Path(__file__).parent / 'commands',
    version = get_version(),
)



from pathlib import Path
from typing import List, Optional

import yaml


def create_config(
    palm_dir,
    image_name,
    plugins: Optional[List] = [],
    protected_branches: Optional[List] = [],
):
    config_path = f"{palm_dir}/config.yaml"
    base_config = {
        "image_name": image_name,
        "plugins": plugins,
        "protected_branches": protected_branches,
    }

    with open(config_path, "w") as file:
        yaml.safe_dump(base_config, file)


def create_command(environment, command: str, template_dir: Path, target_dir: Path):
    """Creates a new cmd file from template"""

    replacements = {
        "command": command,
    }

    environment.generate(template_dir, target_dir, replacements)



from pathlib import Path
from typing import List, Optional

import click

from palm.plugins.core.create_files import *

palm_target_dir = f"{Path.cwd()}/.palm"


@click.group(help="Scaffold new palm commands")
def cli():
    pass


@cli.command()
@click.option(
    "-n",
    "--name",
    multiple=True,
    required=True,
    help="Name of the command(s) you are adding",
)
@click.pass_obj
def command(environment, name: List[str]):
    template_dir = Path(Path(__file__).parents[1], "templates") / 'command'
    """Add a new palm command to the current repo"""
    for command in name:
        create_command(environment, command, template_dir, palm_target_dir)
        click.secho(f"{command} command created in {palm_target_dir}", fg="green")


@cli.command()
@click.option(
    "--group", multiple=False, required=True, help="Name of the command group"
)
@click.option(
    "-c",
    "--command",
    multiple=True,
    required=True,
    help="Name of the commands within the command group",
)
@click.pass_obj
def group(environment, group: str, command: List[str]):
    """Add a new palm command group to the current repo"""
    template_path = Path(Path(__file__).parents[1], "templates") / "command_group"
    replacements = {
        "group": group,
        "commands": command,
    }

    environment.generate(template_path, palm_target_dir, replacements)
    click.secho(f"{group} command group created in {palm_target_dir}", fg="green")


@cli.command("config")
@click.option("-i", "--image-name", multiple=False, help="Name of your docker image")
@click.option("-p", "--plugins", multiple=True, help="List of plugins you want to use")
@click.option(
    "-pb",
    "--protected-branches",
    multiple=True,
    help="List of branches you do not want to run palm on",
)
@click.pass_obj
def config(
    environment,
    image_name: Optional[str],
    plugins: Optional[tuple],
    protected_branches: Optional[tuple],
):
    """Generate a base .palm/config for existing projects"""
    image_name = image_name or environment.palm.image_name
    create_config(palm_target_dir, image_name, plugins, protected_branches)
    click.secho("Palm config created!", fg="green")



import click
import sys
from pathlib import Path


@click.command("new")
@click.option(
    '--global-template', '-g', help='Name of a global cookecutter template to use'
)
@click.option('--template', '-t', help='Cookie cutter template to use')
@click.option('--list', '-l', is_flag=True, help='List global cookiecutter templates')
@click.pass_obj
def cli(environment, global_template: str, template: str, list: bool = False):
    """
    Create a new project with cookiecutter.

    Cookiecutter is a command-line utility that creates projects from
    cookiecutters (project templates).

    This command will create a new project, in a new directory. You should
    run this command from the directory where you want the new project to
    be created.

    Global templates are defined in your global palm config file.
    See the docs for more information:
    https://palm-cli.readthedocs.io/en/latest/code-generation.html#new-projects-with-cookiecutter
    """
    if list:
        list_templates(environment)
        sys.exit(0)

    if global_template:
        default_cookiecutters = environment.palm.config.get('default_cookiecutters', {})
        template = default_cookiecutters.get(global_template)

    if not template:
        click.secho(f'No template defined for type {global_template}', fg='red')
        template = click.prompt('Please enter a cookiecutter template url')

    click.secho(f'Creating new {template} project', fg='green')

    ex_code, _, _ = environment.run_on_host(f'cookiecutter {template}')

    if ex_code == 0:
        click.secho(f'Project created!', fg='green')
    else:
        click.secho(f'Failed to create project', fg='red')
        sys.exit(1)


def list_templates(environment):
    """List global cookiecutter templates"""
    default_cookiecutters = environment.palm.config.get('default_cookiecutters', {})
    click.secho('Global cookiecutter templates:', fg='green')
    for key in default_cookiecutters.keys():
        click.echo(f'  {key}: {default_cookiecutters[key]}')



from pathlib import Path

import yaml
from jinja2 import Environment, FileSystemLoader


class CodeGenerator:
    def __init__(
        self, template_path: Path, target_path: Path, replacements: dict
    ) -> None:
        """Palm core code generator class

        Args:
            template_path (Path): Path to the directory containing your templates and template-config.yaml
            target_path (Path): Path to directory where generated files will be written
            replacements (dict): Dictionary of templated strings to replace
        """
        self.template_path = template_path
        self.target_path = target_path
        self.replacements = replacements
        self.config = self.get_config()

    def run(self) -> str:
        """Runs code generator

        Using the config from template_path/template-config.yaml
        Generate all directories then generate all files with jinja templating.

        Returns:
            str: Result message
        """
        env = Environment(
            loader=FileSystemLoader(self.template_path),
            trim_blocks=True,
            lstrip_blocks=True,
        )

        for directory in self.config.get("directories", []):
            directory_path = Path(
                Path.cwd(),
                self.target_path,
                env.from_string(directory).render(self.replacements),
            )

            if not directory_path.is_dir():
                directory_path.mkdir(parents=True)
            else:
                print(f"{directory_path} already exists")

        for file_item in self.config.get("files", []):
            for key in file_item:
                template = key
                destination = env.from_string(file_item[key]).render(self.replacements)
                print(f"Generating {template} to {destination}")
                t = env.get_template(template)
                templated_contents = t.render(self.replacements)

                with open(Path(Path.cwd(), self.target_path, destination), "w") as fh:
                    fh.write(templated_contents)

        return "Generated successfully"

    def get_config(self) -> dict:
        """Read configuration from template_path/template-config.yaml

        Example template-config.yaml:
            directories:
            - "foo"
            - "foo/{{some_replacement}}"

            files:
            - template_name.py: "path/to/destination.ext"
            - main.py: "foo/{{some_replacement}}/__main__.py"


        Returns:
            dict: dict representation of template-config.yaml
        """
        config_data = {}

        try:
            config_data = yaml.safe_load(
                (self.template_path / "template-config.yaml").read_text()
            )
        except FileNotFoundError:
            print(f"No template-config.yaml found in {self.template_path}")
        except:
            raise Exception("Error reading template-config.yaml")
        return config_data



import click


@click.group()
def cli():
    """{{group}}"""
    pass


{% for command in commands %}
@cli.command()
@click.pass_obj
def {{command}}(environment):
    """{{command}}"""
    click.echo("{{command}} command running!")
    command = "echo {{command}} running" 
    # Run your command in docker:
    # environment.run_in_docker(command)
    # Or on the host:
    # environment.run_on_host(command)

{% endfor %}


import os
from pathlib import Path

import click

from palm.palm_exceptions import AbortPalm


def has_env() -> bool:
    """Check for a .env file in the project root"""
    return Path(".env").exists()


def optionally_create_env():
    """Optionally, create a .env file if it doesn't exist

    Raises:
        AbortPalm: abort if the user doesn't want to create a .env file
    """
    create_env = click.confirm("Would you like palm to create an empty .env file?")
    if create_env:
        Path(".env").touch()
        click.secho(
            "Created an empty .env - please add your ENV vars there", fg="green"
        )
    else:
        raise AbortPalm("Aborting")


def make_executable(path: Path) -> None:
    """Make a file executable

    Args:
        path (Path): Path to file
    """
    mode = os.stat(path).st_mode
    mode |= (mode & 0o444) >> 2  # copy R bits to X
    os.chmod(path, mode)



from pathlib import Path

from palm.plugins.base import BasePlugin

# This plugin is intended for use in testing palm-cli.
# It is not intended to be used in production.

TestInternalPlugin = BasePlugin(
    name="test_internal",
    command_dir=Path(__file__).parent / "commands",
)



import click


@click.command("lint")
@click.pass_obj
def cli(environment):
    """lint the codebase with black"""
    environment.run_in_docker(
        'black --skip-string-normalization --exclude="\.tpl\.py" .'
    )



import importlib
import os
import sys
from typing import Any, Callable, List, Optional

import click
import pkg_resources

from .environment import Environment
from .palm_config import PalmConfig
from .plugin_manager import PluginManager
from .utils import cmd_name_from_file, is_cmd_file, run_on_host

CONTEXT_SETTINGS = dict(auto_envvar_prefix="PALM")

plugin_manager_instance = PluginManager()
palm_config = PalmConfig()


class PalmCLI(click.MultiCommand):
    def __init__(
        self,
        name: Optional[str] = None,
        invoke_without_command: bool = None,
        no_args_is_help: Optional[bool] = None,
        subcommand_metavar: Optional[str] = None,
        chain: bool = None,
        result_callback: Optional[Callable[..., Any]] = None,
        **attrs: Any,
    ) -> None:
        if not palm_config.is_valid_branch():
            msg = f"You are currently on protected branch {palm_config.branch}. For your safety Palm will not run!"
            click.secho(msg, fg="red")
            sys.exit(1)
        self.palm = palm_config
        self.plugin_manager = plugin_manager_instance
        self.plugin_manager.load_plugins(self.palm.plugins)

        super().__init__(
            name=name,
            invoke_without_command=invoke_without_command,
            no_args_is_help=no_args_is_help,
            subcommand_metavar=subcommand_metavar,
            chain=chain,
            result_callback=result_callback,
            **attrs,
        )

    def _commands_from_dir(self, dir) -> List[str]:
        commands = []
        for filename in os.listdir(dir):
            if is_cmd_file(filename):
                commands.append(cmd_name_from_file(filename))
        return commands

    def list_commands(self, ctx) -> List[str]:
        cmds = self.plugin_manager.plugin_command_list
        dedupe = set(cmds)
        cmds = list(dedupe)
        cmds.sort()
        project_excluded_commands = self.palm.config.get("excluded_commands", [])
        cmds = filter(lambda x: x not in project_excluded_commands, cmds)
        return cmds

    def format_commands(self, ctx, formatter) -> None:
        """
        Formats the list of commands for the help page
        Group commands by plugin
        """
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue
            commands.append((subcommand, cmd))

        if len(commands):
            # allow for 3 times the default spacing
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            subsections = {}
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)

                plugin_name = self.plugin_manager.plugin_command_dict.get(cmd.name)
                if not plugin_name:
                    plugin_name = self.plugin_manager.plugin_command_dict.get(
                        subcommand
                    )

                if plugin_name not in subsections:
                    subsections[plugin_name] = []

                subsections[plugin_name].append((subcommand, help))

            for plugin_name, cmds in subsections.items():
                with formatter.section(plugin_name.title()):
                    formatter.write_dl(cmds)

    def get_command(self, ctx, cmd_name: str) -> click.Command:
        try:
            if self.plugin_manager.is_plugin_command(cmd_name):
                spec = self.plugin_manager.command_spec(cmd_name)
            else:
                raise FileNotFoundError
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
        except ImportError as error:
            click.secho(f"Import error: {error}", fg="red")
            return
        except FileNotFoundError:
            click.secho("Command not found, check spelling!", fg="red")
            return
        return mod.cli


def get_version():
    try:
        version = pkg_resources.require("palm")[0].version
    except pkg_resources.DistributionNotFound:
        version = "unknown"
    return version


def required_dependencies_ready():

    docker_checks = (
        (
            "docker --version",
            "Docker is not installed, please install it first",
        ),
        (
            "docker ps",
            "Docker is not running, please start it first",
        ),
    )
    for cmd, msg in docker_checks:
        if run_on_host(cmd, capture_output=True)[0] > 0:
            click.secho(msg, fg="red")
            return False
    return True


@click.group(cls=PalmCLI, context_settings=CONTEXT_SETTINGS)
@click.version_option(get_version())
@click.pass_context
def cli(ctx):
    """Palmetto data product command line interface."""
    is_test = os.getenv("PALM_TEST")
    if not (is_test or required_dependencies_ready()):
        ctx.exit(1)
    ctx.obj = Environment(plugin_manager_instance, palm_config)


if __name__ == "__main__":
    cli()



class AbortPalm(Exception):
    pass


class NoRepositoryError(Exception):
    """
    Raised when a git repository is not found.
    """

    pass


class InvalidConfigError(Exception):
    """
    Raised when the config is invalid.
    """

    def __init__(self, message: str = "Invalid config"):
        self.message = message
        super().__init__(self.message)



from pathlib import Path
from typing import Optional

import click


@click.group(help="Palm plugin utilities")
def cli():
    pass


@cli.command()
@click.option("--name", multiple=False, required=True, help="Name of the plugin")
@click.option("--author", multiple=False, help="Name of the plugin author")
@click.option(
    "--author-email", multiple=False, help="Email address of the plugin author"
)
@click.pass_obj
def new(environment, name: str, author: Optional[str], author_email: Optional[str]):
    """
    Generate a new plugin
    """
    if not author:
        author = click.prompt("What is the author name")

    if not author_email:
        author_email = click.prompt("What is the email address for the author?")

    if name.startswith("palm-"):
        name = name[5:]

    # TODO: Currently, this generator has to be run from withing an existing
    # project. This isn't ideal, but it works. In the future I'd like to move
    # this command out of the core plugin and into a separate
    # system-wide plugin.
    default_target_dir = Path(Path.cwd().parent, f"palm-{name}")
    target_dir = click.prompt(
        "Where do you want to create the plugin?", default=default_target_dir
    )
    target_dir = Path(target_dir)
    if not target_dir.exists():
        click.secho("Target directory does not exist", fg="yellow")
        create_target_dir = click.confirm(
            "Do you want to create the target directory?", default=True
        )
        if not create_target_dir:
            click.secho("Aborting", fg="red")
            return
        target_dir.mkdir(parents=True)

    template_path = Path(Path(__file__).parents[1], "templates") / "plugin"
    replacements = {
        "plugin_name": name,
        "plugin_class_name": f"{name.title().replace('_', '')}Plugin",
        "author": author,
        "author_email": author_email,
    }

    environment.generate(template_path, target_dir, replacements)
    click.secho(f"{name} plugin created in {target_dir}", fg="green")


@cli.command()
@click.option("--name", multiple=False, help="Name of the plugin")
@click.pass_obj
def versions(environment, name: Optional[str]):
    """
    Output plugin versions

    If the name option is provided, output the version for just that plugin
    """

    if name:
        try:
            plugins = [environment.plugin_manager.plugins[name]]
        except KeyError:
            click.secho(f"Plugin {name} not installed in this project", fg="red")
            return
    else:
        plugins = list(environment.plugin_manager.plugins.values())

    # Do not display core or repo 'plugins'
    # since those are versioned with palm core
    excluded_plugins = ["core", "repo"]
    plugins = [p for p in plugins if p.name not in excluded_plugins]

    if not plugins and not name:
        click.secho("No plugins installed", fg="red")
        return

    for plugin in plugins:
        click.echo(f"{plugin.name}: {plugin.version}")


@cli.command()
@click.option("--name", multiple=False, required=True, help="Name of the plugin")
@click.pass_obj
def update(environment, name: Optional[str]):
    """
    Update a plugin

    If the name option is provided, update the specified plugin
    """
    excluded_plugins = ["core", "repo"]
    if name in excluded_plugins:
        click.secho(f"Plugin {name} is a core plugin and cannot be updated", fg="red")

    try:
        plugin = environment.plugin_manager.plugins[name]
    except KeyError:
        click.secho(f"Plugin {name} not installed in this project", fg="red")
        return

    click.echo(f"Updating {plugin.name}...")
    success, message = plugin.update()
    if success:
        click.secho(f"Plugin {plugin.name} updated successfully", fg="green")
    else:
        click.secho(f"Plugin {plugin.name} update failed: {message}", fg="red")


@cli.command()
@click.option("--name", multiple=False, required=True, help="Name of the plugin")
@click.pass_obj
def configure(environment, name: Optional[str]):
    """
    Update a plugin

    If the name option is provided, update the specified plugin
    """
    try:
        plugin = environment.plugin_manager.plugins[name]
    except KeyError:
        click.secho(f"Plugin {name} not installed in this project", fg="red")
        return

    success, _ = plugin.config.update()
    if success:
        click.secho(f"Plugin {plugin.name} configured successfully", fg="green")
    else:
        click.secho(f"Plugin {plugin.name} configuration failed", fg="red")



from pathlib import Path
from typing import Optional

import click

from palm.plugins.core.create_files import *

palm_target_dir = f"{Path.cwd()}/.palm"
templates_dir = Path(Path(__file__).parents[1], "templates").resolve()


@click.command()
@click.option("-i", "--image-name", multiple=False, help="Name of your docker image")
@click.option("-p", "--plugins", multiple=True, help="List of plugins you want to use")
@click.option(
    "-pb",
    "--protected-branches",
    multiple=True,
    help="Branches you do not want to run palm on",
)
@click.option(
    "-c", "--commands", multiple=True, help="List of command names to scaffold"
)
@click.pass_obj
def cli(
    environment,
    image_name: Optional[str],
    plugins: Optional[tuple],
    protected_branches: Optional[tuple],
    commands: Optional[tuple],
):
    """
    Initialize the project for use with Palm

    Creates the .palm directory with __init__.py
    Creates .palm/config.yaml with image_name, plugins, and protected_branches
    based on the options provided
    Creates command files based on the --commands option
    """
    template_dir = Path(Path(__file__).parents[1], "templates") / "command"

    if Path(palm_target_dir).exists():
        click.secho("Palm is already initialized", fg="red")
        return

    Path(".palm").mkdir()
    Path(".palm/__init__.py").touch()

    for command in commands:
        click.echo(f"Adding template for {command}...")
        create_command(environment, command, template_dir, palm_target_dir)

    if not image_name:
        image_name = environment.palm.image_name

    create_config(palm_target_dir, image_name, plugins, protected_branches)

    click.secho("Success! Project initialized with Palm CLI", fg="green")



import click


@click.command('{{command}}')
@click.pass_obj
def cli(environment):
    """{{command}}"""
    click.echo("palm executing {{command}}")
    command = f"echo '{{command}} running!'"
    # Run your command in docker:
    # environment.run_in_docker(command)
    # Or on the host:
    # environment.run_on_host(command)



import shutil
from pathlib import Path

import click


@click.command("override")
@click.option("--name", multiple=False, required=True, help="Name of the command")
@click.pass_obj
def cli(environment, name: str):
    """Override a command in the current project"""
    spec = environment.plugin_manager.command_spec(name)
    origin_path = Path(spec.origin)
    file_name = origin_path.name
    target_path = Path(environment.palm.project_root, ".palm", file_name)

    if target_path.exists():
        click.secho("Command already exists in project, skipping", fg="red")
        return

    if not target_path.parent.exists():
        click.secho(
            "palm is not initialized in this project, please run 'palm init' first",
            fg="red",
        )
        return

    shutil.copy(origin_path, target_path)
    click.secho(f"{name} command overridden to {target_path}", fg="green")



from pathlib import Path
from typing import Optional, List, Union
import yaml
from xmlrpc.client import Boolean
from pygit2 import Repository, discover_repository, GitError

from click import secho, echo
from deepmerge import always_merger
from pygit2 import Repository

from .palm_exceptions import NoRepositoryError


class PalmConfig:
    """Palm config class
    Reads the .palm/config.yaml from the current project
    Makes config available to other modules

    Args:
        project_root: The root path object if not cwd
    """

    project_root: Optional["Path"]
    config: dict = {}
    repo: Optional[Repository] = None
    branch: str = None
    plugins: List[str] = []

    def __init__(self, project_path: Optional["Path"] = Path.cwd()):
        self.project_root = project_path
        self._setup()

    def _setup(self):
        """Setup the config"""
        self.config = self._get_config()
        try:
            self.repo = self._get_repo()
        except NoRepositoryError:
            secho('No git repository found, running in global mode', fg='yellow')
            self.repo = None
        if self.repo:
            self.branch = self._get_current_branch()
            self._use_repo_plugins()
        else:
            self._use_global_plugins()

    def _get_repo(self) -> Repository:
        """Gets the repo object.

        Returns:
            Repository: repo object
        """
        path = discover_repository(str(self.project_root))

        if not path:
            raise NoRepositoryError("No git repository found in the current directory")

        return Repository(path)

    def _get_current_branch(self) -> Union[str, None]:
        """Gets the current branch name.

        Returns:
            Union[str, None]: branch name, or None if not in a repo
        """
        if self.repo:
            try:
                return self.repo.head.shorthand
            except GitError as e:
                secho('Error finding an active branch.', fg='red')
                echo(
                    'If this is a new repo, you should make an initial commit before using palm.\n'
                )
                raise e

        return None

    def _get_config(self) -> dict:
        """Gets both global and repo configs, merging them together.

        Returns:
            dict: dict of merged global and repo configs
        """
        return always_merger.merge(self._get_global_config(), self._get_repo_config())

    def _get_repo_config(self) -> dict:
        """Gets the repo config, reading yaml and returning a dict.
        If the config does not exist, prompt the user to create it.

        Returns:
            dict: dict of repo config, or empty dict if no config
        """
        config_path = self.project_root / ".palm" / "config.yaml"
        if not config_path.exists():
            secho(
                "No palm config found in .palm/config.yml, please run 'palm scaffold config'",
                fg="yellow",
            )
            secho(
                "Some palm commands may not work correctly without palm config \n",
                fg="yellow",
            )
            return {}

        return yaml.safe_load(config_path.read_text())

    def _get_global_config(self, global_config_path: Optional[Path] = None) -> object:
        """Gets the global config, reading yaml and returning a dict.
        If the config does not exist, create it.

        Args:
            global_config_path (Optional[Path], optional): Used for testing. Defaults to None.

        Returns:
            object: dict of global config
        """
        config_path = global_config_path or Path().home() / ".palm" / "config.yaml"
        if not config_path.exists():
            self._create_global_config_file(config_path)

        return yaml.safe_load(config_path.read_text())

    def _create_global_config_file(self, config_path) -> None:
        """Creates the global config file."""
        config_path.parent.mkdir(parents=True, exist_ok=True)
        default_config = {
            'plugins': [],
            'excluded_commands': [],
            'default_cookiecutters': {},
        }
        config_path.write_text(yaml.dump(default_config))

    def _use_repo_plugins(self):
        """Use the default plugins - core, plugins, and repo defined commands"""
        core_plugins = ['core']
        plugins_from_config = self.config.get('plugins') or []
        # The order here defines the order in which commands will be overridden
        # Plugins on the right will override plugins on the left!
        self.plugins = core_plugins + plugins_from_config + ['repo']

    def _use_global_plugins(self):
        """Use the setup plugins - when palm is used outside of a git repo"""
        global_plugins = ['setup']
        plugins_from_config = self.config.get('plugins') or []

        self.plugins = global_plugins + plugins_from_config

    def is_valid_branch(self) -> Boolean:
        """Validate the current branch against the config

        Return:
            Boolean: True if the branch is valid, False if not
        """
        if self.repo and self.branch in self.protected_branches:
            return False

        return True

    @property
    def protected_branches(self) -> List[Optional[str]]:
        """Returns the list of configured protected branches for the current repo

        Returns:
             list[Optional[str]]: list of branch names e.g ['main', 'master']
        """
        return self.config.get("protected_branches") or []

    @property
    def project_root_snake_case(self):
        return self.project_root.name.replace("-", "_")

    @property
    def image_name(self) -> str:
        """Docker image name for the current project
        Attempts to load the image_name from .palm/config.yaml, falling back to
        the snake_cased project root dir name

        Returns:
            str: Name of docker image to use
        """
        return self.config.get('image_name') or self.project_root_snake_case



import sys
from distutils.core import setup
from pathlib import Path

# require version of setuptools that supports find_namespace_packages
from setuptools import setup

try:
    from setuptools import find_namespace_packages
except ImportError:
    # the user has a downlevel version of setuptools.
    print("Error: palm requires setuptools v40.1.0 or higher.")
    print(
        'Please upgrade setuptools with "pip install --upgrade setuptools" and try again'
    )
    sys.exit(1)

# User-friendly description from README.md
this_directory = Path(__file__).parent
long_description = Path(this_directory, "README.md").read_text()

setup(
    name="palm",
    version="2.6.0",  # Don't forget to bump the version in docs/source/conf.py too!
    description="Palm CLI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Palmetto - Data & Analytics team",
    author_email="data-analytics-team@palmetto.com",
    url="https://github.com/palmetto/palm-cli",
    # Packages to include into the distribution
    packages=find_namespace_packages(include=["palm", "palm*"]),
    package_data={"": ["*.yaml", "*.txt"]},
    entry_points="""
    [console_scripts]
    palm=palm.cli:cli
  """,
    license="Apache License 2.0",
    install_requires=Path("requirements.txt").read_text().splitlines(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
    ],
)



import subprocess
from sys import version_info
from typing import Optional, Tuple, List

import click


class UnsupportedVersion(Exception):
    pass


def is_cmd_file(filename: str) -> bool:
    return filename.endswith(".py") and filename.startswith("cmd")


def cmd_name_from_file(filename: str) -> str:
    return filename[4:-3]


def run_on_host(
    cmd: str,
    check: Optional[bool] = False,
    capture_output: Optional[bool] = False,
) -> Tuple[int, str, str]:
    """A simplifed, platform-and-version agnostic interface
    for subprocess.
    By default run_on_host makes some strong
    conventional choices:
     - calls are run as a shell
     - calls are blocking
     - errors in the shell are NOT bubbled up by default
    Args:
        cmd: the command to run.
        check: if True raise a CalledProcessError
        capture_output: if True, stdout and stderr will be suppressed
    Returns:
        Tuple: status code, stdout, stderr

    """
    major, minor, patch, _, _ = version_info
    kwargs = dict(shell=True, check=check)
    if major < 3 or minor < 5:
        raise UnsupportedVersion(f"Python version {major}.{minor} is not supported.")
    if minor < 7:
        from subprocess import PIPE

        if capture_output:
            kwargs.update(dict(stdout=PIPE, stderr=PIPE))
    else:
        kwargs.update(dict(capture_output=capture_output))

    completed = subprocess.run(cmd, **kwargs)
    return (
        completed.returncode,
        completed.stdout.decode("utf-8") if completed.stdout else "",
        completed.stderr.decode("utf-8") if completed.stderr else "",
    )


def run_in_docker(
    cmd: str,
    image_name: str,
    env_vars: Optional[List[str]] = [],
    no_bin_bash: Optional[bool] = False,
    capture_output: Optional[bool] = False,
    silent: Optional[bool] = False,
) -> Tuple[bool, str]:
    """Shells out and runs the cmd in docker

    Args:
        cmd (str): The command you want to run
        env_vars (Optional[dict], optional): Dict of env vars to pass to the docker container.
    """
    if not silent:
        click.secho(f"Executing command `{cmd}` in compose...", fg="yellow")

    docker_cmd = ["docker compose run --service-ports --rm"]
    docker_cmd.extend(env_vars)
    docker_cmd.append(image_name)
    if no_bin_bash:
        docker_cmd.append(cmd)
    else:
        docker_cmd.append(f'/bin/bash -c "{cmd}" ')

    ex_code, std_out, std_err = run_on_host(" ".join(docker_cmd), False, capture_output)
    if capture_output:
        if ex_code == 0:
            return (True, std_out)
        return (False, std_err)

    if ex_code == 0:
        return (True, "Success! Palm completed with exit code 0")
    return (False, f"Fail! Palm exited with code {ex_code}")



import importlib
from pathlib import Path
from typing import List, Optional, Tuple
from urllib.parse import urlparse

from palm.utils import cmd_name_from_file, is_cmd_file, run_on_host
from palm.plugins.base_plugin_config import BasePluginConfig


class BasePlugin:
    def __init__(
        self,
        name: str,
        command_dir: Path,
        version: Optional[str] = "unknown",
        package_location: Optional[str] = None,
        config: Optional[BasePluginConfig] = None,
        **kwargs,
    ) -> None:
        """Initialize a plugin.

        Args:
            name (str): Name of the plugin
            command_dir (Path): Path to the directory containing commands
            version (Optional[str], optional): Plugin version. Defaults to "unknown".
            package_location (Optional[str], optional): Either pypi package name
                or github https URL. Used for installation Defaults to None.
        """
        self.name = name
        self.command_dir = command_dir
        self.version = version
        self.package_location = package_location
        self.__dict__.update(kwargs)
        self.config = config

    def all_commands(self) -> List:
        """Get the commands for a plugin.

        Returns:
            list: List of click cli command names
        """
        command_list = []

        for path in self.command_dir.glob("*.py"):
            if is_cmd_file(path.name):
                command_list.append(cmd_name_from_file(path.name))

        return command_list

    def command_map(self) -> dict:
        """Returns a dict that maps all commands to the plugin name.
        Used by PluginManager to identify which plugin a command belongs to.

        Returns:
            dict: dict where key is command name and value is plugin name
        """
        return dict.fromkeys(self.all_commands(), self.name)

    def get_command(self, command_name: str) -> importlib.machinery.ModuleSpec:
        """Get the modulespec for a given command

        Args:
            command_name (str): Name of the command

        Returns:
            importlib.ModuleSpec: ModuleSpec for the command
        """
        command_path = self.command_dir / f"cmd_{command_name}.py"
        return importlib.util.spec_from_file_location(command_name, command_path)

    def update(self) -> Tuple[bool, str]:
        """Update the plugin.

        Returns:
            Tuple[bool, str]: Tuple of ('success', 'message')
        """
        if self.package_location is None:
            return (False, "This plugin does not support upgrading via palm")

        package_url = urlparse(self.package_location)
        if package_url.hostname == "github.com":
            install_url = f"git+{self.package_location}"
        else:  # This should be a pypi package name
            install_url = self.package_location

        # TODO: do a better job of the package name, prepend 'palm-' isn't great.
        uninstall_cmd = f"python3 -m pip uninstall -y palm-{self.name}"
        upgrade_cmd = f"python3 -m pip install {install_url}"

        for cmd in (uninstall_cmd, upgrade_cmd):
            _, _, _ = run_on_host(cmd, check=True)
        return (True, "Plugin upgraded successfully")



import shlex
import subprocess

import click


@click.command("update")
@click.option("--local", is_flag=True, help="attempts to install from cwd")
def cli(local: bool):
    """This updates the current version of palm. meta huh?"""

    if not click.confirm(
        f"\nAttempt to update Palm {'from cwd' if local else 'from github'}?"
    ):
        click.secho("Cancelled", fg="red")
        return

    uninstall_cmd = "python3 -m pip uninstall -y palm"
    if local:
        upgrade_cmd = f"python3 -m pip install ."
    else:
        upgrade_cmd = f"pip install palm"

    for cmd in (
        uninstall_cmd,
        upgrade_cmd,
    ):
        subprocess.run(shlex.split(cmd), check=True)

    click.secho("Success! Palm has been updated.", fg="green")



import click
import yaml
from typing import Tuple
from abc import ABC, abstractmethod
from pathlib import Path
from pydantic import BaseModel, ValidationError

from palm.palm_exceptions import InvalidConfigError


class BasePluginConfig(ABC):
    def __init__(self, plugin_name: str, model: BaseModel):
        self.plugin_name = plugin_name
        self.config_path = Path.cwd() / ".palm" / f"config.yaml"
        self.model = model

    @abstractmethod
    def set(self) -> dict:
        """Setter for plugin config

        This method should be implemented by the plugin to set the config
        for the plugin. It should return a dict of the config that was set.

        Returns:
            dict: The config that was set
        """
        pass

    def update(self) -> Tuple[bool, dict]:
        """Updates the config file with configuration provided by the set method

        Returns:
            Tuple[bool, dict]: Returns a tuple of (True, config) if the config was
            updated successfully, and (False, {}) if it was not.
        """
        try:
            config = self.set()
            self.validate(config)
            self._write(config)
        except InvalidConfigError as e:
            click.secho(str(e), fg="red")
            raise e
        except Exception as e:
            click.secho(str(e), fg="red")
            return (False, {})
        return (True, config)

    def get(self) -> BaseModel:
        if not self.config_path.exists():
            click.secho(
                f"Config file not found at {self.config_path}, run palm init", fg="red"
            )
            return {}

        config_dict = self._read()
        return self.model(**config_dict)

    def validate(self, config: dict) -> bool:
        """Validates the plugin config against the pydantic model

        Args:
            config (dict): The config to validate

        Returns:
            bool: Returns True if the config is valid, Raises InvalidConfigError if not
        """
        try:
            self.model(**config)
            return True
        except ValidationError as e:
            msg = f"Invalid config for plugin {self.plugin_name}: {e}"
            raise InvalidConfigError(msg)

    def _read(self) -> dict:
        palm_config = yaml.load(self.config_path.read_text(), Loader=yaml.FullLoader)
        plugin_config = palm_config.get('plugin_config', {}).get(self.plugin_name, {})

        if not plugin_config:
            _, plugin_config = self.update()

        return plugin_config

    def _write(self, config: dict):
        if not self.config_path.exists():
            click.secho(
                f"Config file not found at {self.config_path}, run palm init", fg="red"
            )
            return

        palm_config = yaml.load(self.config_path.read_text(), Loader=yaml.FullLoader)
        if not 'plugin_config' in palm_config.keys():
            palm_config['plugin_config'] = {}

        if not self.validate(config):
            click.secho("Invalid config, not writing to config file", fg="red")
            raise InvalidConfigError

        palm_config['plugin_config'][self.plugin_name] = config
        self.config_path.write_text(yaml.dump(palm_config))
